from typing import Any, Self

from sqlalchemy.orm import DeclarativeBase as OriginalBase


class DeclarativeBase(OriginalBase):
    __abstract__ = True
    __mapper_args__ = {"eager_defaults": True}

    @classmethod
    def _get_primary_key(cls) -> list[str]:
        primary_key = [c.name for c in cls.__mapper__.primary_key]
        return primary_key

    @classmethod
    def _get_attributes(cls) -> list[str]:
        primary_key = set(cls._get_primary_key())
        attributes = {c.name for c in cls.__mapper__.columns} | {
            r.key for r in cls.__mapper__.relationships
        }
        safe_attributes = list(attributes - primary_key)
        return safe_attributes

    @classmethod
    def new(cls, **kwargs: Any) -> Self:
        safe_kwargs = {k: v for k, v in kwargs.items() if k in cls._get_attributes()}
        return cls(**safe_kwargs)

    def update(self, **kwargs: Any) -> None:
        safe_kwargs = {k: v for k, v in kwargs.items() if k in self._get_attributes()}
        for k, v in safe_kwargs.items():
            setattr(self, k, v)



from collections.abc import Generator, Sequence
from contextlib import contextmanager
from typing import Any, TypeVar

from sqlalchemy import Engine, func, select
from sqlalchemy.orm import Session

from inzicht.crud.errors import DoesNotExistError
from inzicht.crud.interfaces import CRUDInterface
from inzicht.declarative import DeclarativeBase

T = TypeVar("T", bound=DeclarativeBase)


@contextmanager
def session_factory(bind: Engine) -> Generator[Session, None, None]:
    with Session(bind=bind, expire_on_commit=False) as session:
        try:
            session.begin()
            yield session
        except:
            session.rollback()
            raise
        else:
            session.commit()


class GenericCRUD(CRUDInterface[T]):
    def __init__(self, session: Session) -> None:
        self.session = session

    @classmethod
    def get_model(cls) -> type[T]:
        (bases,) = cls.__orig_bases__  # type: ignore  # noqa
        (model,) = bases.__args__
        return model

    def count(self, where: Any | None = None) -> int:
        model = self.get_model()
        query = select(func.count()).select_from(model)
        if where is not None:
            query = query.filter(where)
        count = self.session.execute(query).scalar() or 0
        return count

    def create(self, *, payload: dict[str, Any]) -> T:
        model = self.get_model()
        instance = model.new(**payload)
        self.session.add(instance)
        self.session.flush()
        return instance

    def create_many(self, *, payload: Sequence[dict[str, Any]]) -> Sequence[T]:
        model = self.get_model()
        instances = [model.new(**item) for item in payload]
        self.session.add_all(instances)
        self.session.flush()
        return instances

    def read(self, id: int | str) -> T:
        model = self.get_model()
        instance = self.session.get(model, id)
        if not instance:
            raise DoesNotExistError(
                f"Instance of model='{model}' with id='{id}' was not found"
            )
        return instance

    def read_many(
        self,
        *,
        where: Any | None = None,
        order_by: Any | None = None,
        skip: int = 0,
        take: int = 10,
    ) -> Generator[T, None, None]:
        model = self.get_model()
        query = select(model)
        if where is not None:
            query = query.filter(where)
        if order_by is not None:
            query = query.order_by(order_by)
        if skip:
            query = query.offset(skip)
        if take:
            query = query.limit(take)
        items = (item for item in self.session.execute(query).scalars())
        return items

    def update(self, id: int | str, *, payload: dict[str, Any]) -> T:
        instance = self.read(id=id)
        instance.update(**payload)
        self.session.add(instance)
        self.session.flush()
        return instance

    def delete(self, id: int | str) -> T:
        instance = self.read(id=id)
        self.session.delete(instance)
        self.session.flush()
        return instance



from abc import ABC, abstractmethod
from collections.abc import Generator, Sequence
from typing import Any, Generic, TypeVar

from inzicht.declarative import DeclarativeBase

T = TypeVar("T", bound=DeclarativeBase)


class CRUDInterface(ABC, Generic[T]):
    """
    Abstract base class that defines the CRUD interface for generic operations on a resource of type T.
    """

    @classmethod
    @abstractmethod
    def get_model(cls) -> type[T]:
        """
        Retrieves the model class associated with the GenericCRUD class.

        Returns:
            type[T]: The model class associated with the GenericCRUD class.
        """

    @abstractmethod
    def count(self, where: Any | None = None) -> int:
        """
        Count the total number of records.

        Args:
            where (Any, optional): Filter conditions for retrieving records.

        Returns:
            int: The total number of records in the collection.
        """

    @abstractmethod
    def create(self, *, payload: dict[str, Any]) -> T:
        """
        Create a new record with the provided payload.

        Args:
            payload (dict[str, Any]): The data to create the new record.

        Returns:
            T: The created record.
        """

    @abstractmethod
    def create_many(self, *, payload: Sequence[dict[str, Any]]) -> Sequence[T]:
        """
        Create multiple records with the provided payload.

        Args:
            payload (Sequence[dict[str, Any]]): A sequence of data dictionaries to create the new records.

        Returns:
            Sequence[T]: A sequence of created records.
        """

    @abstractmethod
    def read(self, id: int | str) -> T:
        """
        Retrieve a single record by its ID.

        Args:
            id (int | str): The ID of the record to retrieve.

        Returns:
            T: The record with the specified ID.
        """

    @abstractmethod
    def read_many(
        self,
        *,
        where: Any | None = None,
        order_by: Any | None = None,
        skip: int = 0,
        take: int = 10,
    ) -> Generator[T, None, None]:
        """
        Retrieve multiple records based on conditions.

        Args:
            where (Any, optional): Filter conditions for retrieving records.
            order_by (Any, optional): Criteria to order the results.
            skip (int, optional): Number of records to skip. Defaults to 0.
            take (int, optional): Number of records to retrieve. Defaults to 10.

        Returns:
            Generator[T, None, None]: A generator of the retrieved records.
        """

    @abstractmethod
    def update(self, id: int | str, *, payload: dict[str, Any]) -> T:
        """
        Update a record by its ID with the provided payload.

        Args:
            id (int | str): The ID of the record to update.
            payload (dict[str, Any]): The data to update the record with.

        Returns:
            T: The updated record.
        """

    @abstractmethod
    def delete(self, id: int | str) -> T:
        """
        Delete a record by its ID.

        Args:
            id (int | str): The ID of the record to delete.

        Returns:
            T: The deleted record.
        """
